@@ -1382,6 +1382,17 @@ class LUT3DFrame(BaseFrame):
 					   getcfg("3dlut.trc_gamma.backup"))
 				setcfg("3dlut.trc_gamma.backup", None)
 			self.lut3d_update_trc_controls()
+		elif (hasattr(self, "input_profile") and
+			  isinstance(self.input_profile.tags.get("A2B0"),
+						 ICCP.LUT16Type) and
+			  isinstance(self.input_profile.tags.get("A2B1", ICCP.LUT16Type()),
+						 ICCP.LUT16Type) and
+			  isinstance(self.input_profile.tags.get("A2B2", ICCP.LUT16Type()),
+						 ICCP.LUT16Type)):
+			self.lut3d_trc_apply_black_offset_ctrl.Enable()
+		else:
+			self.lut3d_trc_apply_black_offset_ctrl.Disable()
+			setcfg("3dlut.apply_black_offset", 0)
 		if getcfg("3dlut.apply_black_offset"):
 			self.lut3d_trc_apply_black_offset_ctrl.SetValue(True)
 		elif getcfg("3dlut.apply_trc"):
